from .fragment_growth import enumerate_one_step_grown_fragments
from .is_new import is_new
from .key import get_key
from .starting_fragments import get_unique_starting_fragments
from .sub_fragment import create_sub_fragment
from .symmetry_operation import validate_symmetry_operation
from .unique import extract_unique_fragments



from nasap_net.io.assemblies.semi_light_assembly import SemiLightAssembly
from nasap_net.models import Bond


def test___repr__():
    assembly = SemiLightAssembly(
        components={'X0': 'X'},
        bonds=[],
    )
    repr_str = repr(assembly)
    expected_str = (
        "<SemiLightAssembly components={'X0': 'X'}, bonds=[]>"
    )
    assert repr_str == expected_str


def test___repr___with_id():
    assembly = SemiLightAssembly(
        id_='free_X',
        components={'X0': 'X'},
        bonds=[],
    )
    repr_str = repr(assembly)
    expected_str = (
        "<SemiLightAssembly id_='free_X', components={'X0': 'X'}, bonds=[]>"
    )
    assert repr_str == expected_str


def test___repr___with_bonds():
    assembly = SemiLightAssembly(
        components={'X0': 'X', 'M0': 'M', 'X1': 'X'},
        bonds=[
            Bond('X0', 0, 'M0', 0),
            Bond('M0', 1, 'X1', 0),
        ],
    )
    repr_str = repr(assembly)
    expected_str = (
        "<SemiLightAssembly components={'X0': 'X', 'M0': 'M', 'X1': 'X'}, "
        "bonds=[('M0', 0, 'X0', 0), ('M0', 1, 'X1', 0)]>"
    )
    assert repr_str == expected_str



from collections import defaultdict, deque
from collections.abc import Hashable, Iterable, Mapping
from typing import Any

from nasap_net.models import Assembly
from nasap_net.types import ID
from .lib import enumerate_one_step_grown_fragments, get_key, \
    get_unique_starting_fragments, is_new, validate_symmetry_operation
from .models import Fragment
from .models.fragment import create_complete_fragment


def enumerate_fragments(
        template: Assembly,
        symmetry_operations: Iterable[Mapping[Any, ID]] | None = None
) -> set[Assembly]:
    template_fragment = create_complete_fragment(template)
    if symmetry_operations is not None:
        for sym_op in symmetry_operations:
            validate_symmetry_operation(template_fragment, sym_op)

    found: defaultdict[Hashable, set[Fragment]] = defaultdict(set)

    starting_fragments = set(get_unique_starting_fragments(
        template_fragment, symmetry_operations)
    )
    for frag in starting_fragments:
        found[get_key(frag)].add(frag)
    queue = deque(sorted(starting_fragments))

    while queue:
        cur_frag = queue.popleft()
        one_step_grown_frags = enumerate_one_step_grown_fragments(cur_frag)
        for frag in one_step_grown_frags:
            if is_new(frag, found, symmetry_operations):
                found[get_key(frag)].add(frag)
                queue.append(frag)

    result: set[Assembly] = set()
    for frags in found.values():
        result.update(frag.to_assembly() for frag in frags)
    return result



from typing import Any

import yaml

from nasap_net.io.assemblies.semi_light_assembly import SemiLightAssembly
from nasap_net.models import Bond


def dump_semi_light_assemblies(assemblies: Any) -> str:
    """Dump light assemblies to a YAML string."""
    return yaml.dump(
        assemblies,
        Dumper=_SemiLightAssemblyDumper,
        sort_keys=False,
        default_flow_style=None,
    )


class _SemiLightAssemblyDumper(yaml.SafeDumper):
    def ignore_aliases(self, _):
        return True


def _semi_light_assembly_representer(
        dumper: _SemiLightAssemblyDumper, data: SemiLightAssembly
) -> yaml.MappingNode:
    mapping: dict = {
        'components': dict(sorted(data.components.items())),
        'bonds': sorted(data.bonds),
    }
    if data.id_or_none is not None:
        mapping['id_'] = data.id_
    return dumper.represent_mapping('!Assembly', mapping)


def _bond_representer(
        dumper: _SemiLightAssemblyDumper, data: Bond
) -> yaml.SequenceNode:
    site1, site2 = sorted(data.sites)
    return dumper.represent_list([
        site1.component_id,
        site1.site_id,
        site2.component_id,
        site2.site_id,
    ])


yaml.add_representer(
    SemiLightAssembly,
    _semi_light_assembly_representer,
    Dumper=_SemiLightAssemblyDumper,
)
yaml.add_representer(Bond, _bond_representer, Dumper=_SemiLightAssemblyDumper)



from __future__ import annotations

from dataclasses import dataclass
from functools import total_ordering
from typing import Iterable, Iterator, Self

from nasap_net.models import Assembly
from nasap_net.types import ID
from .growing_step import GrowingStep
from .light_bond import LightBond
from .template_assembly import TemplateAssembly


@total_ordering
@dataclass(frozen=True, init=False)
class Fragment:
    components: frozenset[ID]
    bonds: frozenset[LightBond]
    template: TemplateAssembly

    def __init__(
            self,
            components: Iterable[ID],
            bonds: Iterable[LightBond],
            template: Assembly,
    ) -> None:
        object.__setattr__(self, 'components', frozenset(components))
        object.__setattr__(self, 'bonds', frozenset(bonds))
        object.__setattr__(
            self,
            'template',
            TemplateAssembly(assembly=template),
        )
        self._validate()

    def __lt__(self, other):
        if not isinstance(other, Fragment):
            return NotImplemented
        if self.components != other.components:
            return sorted(self.components) < sorted(other.components)
        return sorted(self.bonds) < sorted(other.bonds)

    def enumerate_possible_growing_steps(self) -> Iterator[GrowingStep]:
        """Return possible growing steps that add bonds between existing
        components.
        """
        possible_bonds = self.template.get_bonds_involving_components(
            self.components
        ) - self.bonds
        for bond in sorted(possible_bonds):
            missing_comps = bond.component_ids - self.components
            if len(missing_comps) == 0:
                # Type 1 growing step: bond between existing components
                yield GrowingStep(bond_to_add=bond, component_to_add=None)
            elif len(missing_comps) == 1:
                # Type 2 growing step: bond involving one new component
                new_comp_id = list(missing_comps)[0]
                yield GrowingStep(
                    bond_to_add=bond, component_to_add=new_comp_id
                )

    def copy_with(
            self,
            components: Iterable[ID] | None = None,
            bonds: Iterable[LightBond] | None = None,
    ) -> Self:
        return self.__class__(
            components=(
                components if components is not None else self.components
            ),
            bonds=bonds if bonds is not None else self.bonds,
            template=self.template.assembly,
        )

    def to_assembly(self) -> Assembly:
        """Convert the fragment back to an Assembly.

        Returns
        -------
        Assembly
            The Assembly representation of the Fragment.

        Notes
        -----
        The conversion from Fragment to Assembly assumes that
        a pair of component IDs uniquely identifies a bond,
        which holds true as long as there are no parallel bonds (chelate).
        This logic needs to be revisited if parallel bonds are to be supported.
        """
        components = {
            comp_id: self.template.assembly.components[comp_id]
            for comp_id in self.components
        }
        bonds = [
            self.template.assembly.get_bond_by_comp_ids(*bond.component_ids)
            for bond in self.bonds
        ]
        return Assembly(components=components, bonds=bonds)

    def _validate(self):
        for bond in self.bonds:
            for comp_id in bond.component_ids:
                if comp_id not in self.components:
                    raise ValueError(
                        f"Bond references non-existent component ID: {comp_id}"
                    )


def create_complete_fragment(assembly: Assembly) -> Fragment:
    """Create a complete fragment from an assembly."""
    components = assembly.components.keys()
    bonds = [LightBond.from_bond(bond) for bond in assembly.bonds]
    return Fragment(components=components, bonds=bonds, template=assembly)



from nasap_net.helpers import assign_composition_formula_ids
from nasap_net.models import Assembly, Component


def test():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])

    MX2 = Assembly(
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[],
    )
    MX2_2 = Assembly(
        components={'X0_2': X, 'M0_2': M, 'X1_2': X},
        bonds=[],
    )

    assert assign_composition_formula_ids([MX2, MX2_2]) == [
        Assembly(
            id_='MX2',
            components={'X0': X, 'M0': M, 'X1': X},
            bonds=[],
        ),
        Assembly(
            id_='MX2_2',
            components={'X0_2': X, 'M0_2': M, 'X1_2': X},
            bonds=[],
        ),
    ]


def test_order():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])

    MX2 = Assembly(
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[],
    )

    assert assign_composition_formula_ids(
        [MX2], order=['X', 'M']) == [
        Assembly(
            id_='X2M',
            components={'X0': X, 'M0': M, 'X1': X},
            bonds=[],
        ),
    ]



from nasap_net.assembly_enumeration import enumerate_assemblies


def test_M4L4(M4L4, M4L4_symmetry_operations, X):
    assemblies = enumerate_assemblies(
        M4L4,
        leaving_ligand=X,
        leaving_ligand_site=0,
        metal_kinds=['M'],
        symmetry_operations=list(M4L4_symmetry_operations.values())
    )
    assert len(assemblies) == 14


def test_M2L4(M2L4, M2L4_symmetry_operations, X):
    assemblies = enumerate_assemblies(
        M2L4,
        leaving_ligand=X,
        leaving_ligand_site=0,
        metal_kinds=['M'],
        symmetry_operations=list(M2L4_symmetry_operations.values())
    )
    assert len(assemblies) == 29


def test_M9L6(M9L6, M9L6_symmetry_operations, X):
    assemblies = enumerate_assemblies(
        M9L6,
        leaving_ligand=X,
        leaving_ligand_site=0,
        metal_kinds=['M'],
        symmetry_operations=list(M9L6_symmetry_operations.values())
    )
    assert len(assemblies) == 505


def test_without_specifying_leaving_site(M4L4, M4L4_symmetry_operations, X):
    assemblies = enumerate_assemblies(
        M4L4,
        leaving_ligand=X,
        leaving_ligand_site=None,
        metal_kinds=['M'],
        symmetry_operations=list(M4L4_symmetry_operations.values())
    )
    assert len(assemblies) == 14



from typing import Hashable

from nasap_net.models import Assembly


def get_light_signature(assembly: Assembly) -> Hashable:
    """Compute a light signature of the assembly for quick filtering.

    Assemblies with different signatures are guaranteed to be non-isomorphic.
    Assemblies with the same signature may or may not be isomorphic.

    The signature consists of:
    - A sorted tuple of component kinds.
    - A sorted tuple of sorted tuples of bond component kinds.

    Parameters
    ----------
    assembly : Assembly
        The assembly to compute the signature for.

    Returns
    -------
    Hashable
        The light signature of the assembly.
    """
    return (
        # component kinds
        tuple(sorted(comp.kind for comp in assembly.components.values())),
        # bond component kinds
        tuple(sorted(
            tuple(sorted([
                assembly.get_component_kind_of_site(site1),
                assembly.get_component_kind_of_site(site2),
            ]))
            for site1, site2 in assembly.bonds
        ))
    )



from .rich_to_semi_light import ConversionResult, \
    convert_assemblies_to_semi_light_ones
from .semi_light_assembly import SemiLightAssembly
from .semi_light_to_rich import convert_semi_light_assemblies_to_rich_ones, \
    convert_semi_light_assembly_to_rich_one



import logging
import os
from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import Literal

import pandas as pd

from nasap_net.helpers import validate_unique_assembly_ids
from nasap_net.models import Assembly, BindingSite, Reaction
from nasap_net.types import ID
from .models import ReactionRow

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def load_reactions(
        file_path: os.PathLike | str,
        assemblies: Iterable[Assembly],
        *,
        assembly_id_type: Literal['str', 'int'] = 'str',
        component_id_type: Literal['str', 'int'] = 'str',
        site_id_type: Literal['str', 'int'] = 'str',
        reaction_id_type: Literal['str', 'int'] = 'str',
        has_index_column: bool = False,
) -> list[Reaction]:
    """Load reactions from a CSV file and convert to Reaction objects."""
    validate_unique_assembly_ids(assemblies)
    id_to_assembly = {assembly.id_: assembly for assembly in assemblies}

    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f'File "{str(file_path)}" does not exist.')

    df = pd.read_csv(
        file_path,
        index_col=0 if has_index_column else None,
    )
    df = df.astype(object).where(pd.notnull(df), None)

    types = {'int': int, 'str': str}

    reaction_rows = [
        ReactionRow.from_dict(
            row,
            assembly_id_type=types[assembly_id_type],
            component_id_type=types[component_id_type],
            site_id_type=types[site_id_type],
            reaction_id_type=types[reaction_id_type],
        )
        for row in df.to_dict(orient="records")
    ]

    reactions = [
        reaction_row_to_reaction(reaction_row, id_to_assembly)
        for reaction_row in reaction_rows
    ]

    logger.info('Loaded %d reactions from "%s"', len(reactions), str(file_path))

    return reactions


def reaction_row_to_reaction(
        reaction_row: ReactionRow,
        id_to_assemblies: Mapping[ID, Assembly],
) -> Reaction:
    """Convert a ReactionRow to a Reaction object."""
    init_assem = id_to_assemblies[reaction_row.init_assem_id]
    entering_assem = (
        None if reaction_row.entering_assem_id is None
        else id_to_assemblies[reaction_row.entering_assem_id]
    )
    product_assem = id_to_assemblies[reaction_row.product_assem_id]
    leaving_assem = (
        None if reaction_row.leaving_assem_id is None
        else id_to_assemblies[reaction_row.leaving_assem_id]
    )

    return Reaction(
        init_assem=init_assem,
        entering_assem=entering_assem,
        product_assem=product_assem,
        leaving_assem=leaving_assem,
        metal_bs=BindingSite(
            component_id=reaction_row.metal_bs_component,
            site_id=reaction_row.metal_bs_site,
        ),
        leaving_bs=BindingSite(
            component_id=reaction_row.leaving_bs_component,
            site_id=reaction_row.leaving_bs_site,
        ),
        entering_bs=BindingSite(
            component_id=reaction_row.entering_bs_component,
            site_id=reaction_row.entering_bs_site,
        ),
        duplicate_count=reaction_row.duplicate_count,
        id_=reaction_row.id_,
    )



from dataclasses import asdict, dataclass
from typing import Generic, TypeVar

from nasap_net.models import Reaction
from .const import Column

_T = TypeVar('_T', int, str)  # Assembly ID type
_S = TypeVar('_S', int, str)  # Component ID type
_U = TypeVar('_U', int, str)  # Site ID type
_R = TypeVar('_R', int, str)  # Reaction ID type

T = TypeVar('T', bound='ReactionRow')


@dataclass(frozen=True)
class ReactionRow(Generic[_T, _S, _U, _R]):
    init_assem_id: _T
    entering_assem_id: _T | None
    product_assem_id: _T
    leaving_assem_id: _T | None
    metal_bs_component: _S
    metal_bs_site: _U
    leaving_bs_component: _S
    leaving_bs_site: _U
    entering_bs_component: _S
    entering_bs_site: _U
    duplicate_count: int
    id_: _R | None

    def to_dict(self) -> dict:
        return asdict(self)

    @classmethod
    def from_reaction(cls: type[T], reaction: Reaction) -> T:
        return cls(
            init_assem_id=reaction.init_assem_id,
            entering_assem_id=reaction.entering_assem_id,
            product_assem_id=reaction.product_assem_id,
            leaving_assem_id=reaction.leaving_assem_id,
            metal_bs_component=reaction.metal_bs.component_id,
            metal_bs_site=reaction.metal_bs.site_id,
            leaving_bs_component=reaction.leaving_bs.component_id,
            leaving_bs_site=reaction.leaving_bs.site_id,
            entering_bs_component=reaction.entering_bs.component_id,
            entering_bs_site=reaction.entering_bs.site_id,
            duplicate_count=reaction.duplicate_count,
            id_=reaction.id_or_none,
        )

    @classmethod
    def from_dict(
            cls: type[T],
            data: dict,
            *,
            assembly_id_type: type[_T],
            component_id_type: type[_S],
            site_id_type: type[_U],
            reaction_id_type: type[_R],
    ) -> T:
        return cls(
            init_assem_id=assembly_id_type(data[Column.INIT_ASSEM_ID.value]),
            entering_assem_id=(
                assembly_id_type(data[Column.ENTERING_ASSEM_ID.value])
                if data[Column.ENTERING_ASSEM_ID.value] is not None else None
            ),
            product_assem_id=assembly_id_type(
                data[Column.PRODUCT_ASSEM_ID.value]
            ),
            leaving_assem_id=(
                assembly_id_type(data[Column.LEAVING_ASSEM_ID.value])
                if data[Column.LEAVING_ASSEM_ID.value] is not None else None
            ),
            metal_bs_component=component_id_type(
                data[Column.METAL_BS_COMPONENT.value]
            ),
            metal_bs_site=site_id_type(data[Column.METAL_BS_SITE.value]),
            leaving_bs_component=component_id_type(
                data[Column.LEAVING_BS_COMPONENT.value]
            ),
            leaving_bs_site=site_id_type(data[Column.LEAVING_BS_SITE.value]),
            entering_bs_component=component_id_type(
                data[Column.ENTERING_BS_COMPONENT.value]
            ),
            entering_bs_site=site_id_type(data[Column.ENTERING_BS_SITE.value]),
            duplicate_count=int(data[Column.DUPLICATE_COUNT.value]),
            id_=(
                reaction_id_type(data[Column.ID_.value])
                if data.get(Column.ID_.value) is not None else None
            ),
        )



from collections.abc import Iterable

from nasap_net.exceptions import DuplicateIDError, IDNotSetError
from nasap_net.models import Assembly
from nasap_net.types import ID


def validate_unique_assembly_ids(assemblies: Iterable[Assembly]) -> None:
    """Validate that all assemblies have unique IDs.

    Parameters
    ----------
    assemblies : Iterable[Assembly]
        The assemblies to validate.

    Raises
    ------
    IDNotSetError
        If any assembly does not have an ID set.
    DuplicateIDError
        If any duplicate assembly IDs are found.
    """
    assembly_ids: set[ID] = set()
    # All assemblies must have IDs and unique IDs
    for assem in assemblies:
        if assem.id_or_none is None:
            raise IDNotSetError(
                'All assemblies must have IDs.'
            )
        if assem.id_ in assembly_ids:
            raise DuplicateIDError(
                f'Duplicate assembly ID found: {assem.id_}'
            )
        assembly_ids.add(assem.id_)



from .conversion import GraphConversionResult
from ..models import Isomorphism


def decode_mapping(
        mapping: list[int],
        conv_res1: GraphConversionResult,
        conv_res2: GraphConversionResult,
) -> Isomorphism:
    comp_id_mapping = {}
    binding_site_mapping = {}
    for v1, v2 in enumerate(mapping):
        if v1 in conv_res1.core_mapping.inv:
            comp_id1 = conv_res1.core_mapping.inv[v1]
            comp_id2 = conv_res2.core_mapping.inv[v2]
            comp_id_mapping[comp_id1] = comp_id2
        else:
            assert v1 in conv_res1.binding_site_mapping.inv
            site1 = conv_res1.binding_site_mapping.inv[v1]
            site2 = conv_res2.binding_site_mapping.inv[v2]
            binding_site_mapping[site1] = site2
    return Isomorphism(
        comp_id_mapping=comp_id_mapping,
        binding_site_mapping=binding_site_mapping
    )



from __future__ import annotations

from dataclasses import dataclass
from functools import total_ordering
from typing import Self

from nasap_net.models import Bond
from nasap_net.types import ID


@total_ordering
@dataclass(frozen=True, init=False)
class LightBond:
    component_ids: frozenset[ID]

    def __init__(self, comp_id1: ID, comp_id2: ID) -> None:
        object.__setattr__(
            self, 'component_ids', frozenset({comp_id1, comp_id2})
        )

    def __lt__(self, other):
        if not isinstance(other, LightBond):
            return NotImplemented
        return sorted(self.component_ids) < sorted(other.component_ids)

    @classmethod
    def from_bond(cls, bond: Bond) -> Self:
        site1, site2 = bond.sites
        return cls(site1.component_id, site2.component_id)



from nasap_net.utils.default import MISSING, default_if_missing


def test_with_value():
    result = default_if_missing('value', 'default')
    assert result == 'value'


def test_with_missing():
    result = default_if_missing(MISSING, 'default')
    assert result == 'default'


def test_no_fallback_for_none():
    result = default_if_missing(None, 'default')
    assert result is None



from .binding_site_equivalence import UniqueComb, \
    extract_unique_site_combinations, group_equivalent_node_combs
from .reaction_resolver import ReactionOutOfScopeError, ReactionResolver
from .separation import SeparatedIntoMoreThanTwoPartsError, \
    separate_if_possible



import logging
import os
from collections.abc import Iterable
from pathlib import Path

import pandas as pd

from nasap_net.models import Reaction
from .models import ReactionRow

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def save_reactions(
        reactions: Iterable[Reaction],
        file_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        index: bool = False,
        ) -> None:
    """Save reactions to a CSV file.

    Resulting CSV columns:
    - init_assem_id : str | int
    - entering_assem_id : str | int | None
    - product_assem_id : str | int
    - leaving_assem_id : str | int | None
    - metal_bs_component : str | int
    - metal_bs_site : str | int
    - leaving_bs_component : str | int
    - leaving_bs_site : str | int
    - entering_bs_component : str | int
    - entering_bs_site : str | int
    - duplicate_count : int
    - id : str | int | None

    Parameters
    ----------
    reactions : Iterable[Reaction]
        Reactions to save.
    file_path : os.PathLike | str
        Path to the CSV file to write.
    overwrite : bool, optional
        If True, overwrite the file if it already exists.
        If False, raise an error if the file already exists.
        Default is False.
    index : bool, optional
        If True, write row names (index). Default is False.
    """
    file_path = Path(file_path)
    if file_path.exists() and not overwrite:
        raise FileExistsError(
            f'File "{str(file_path)}" already exists. '
            'Use `overwrite=True` to overwrite it.'
        )

    df = reactions_to_df(reactions)

    file_path.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(file_path, index=index)
    logger.info('Saved! --> "%s"', str(file_path))


def reactions_to_df(reactions: Iterable[Reaction]) -> pd.DataFrame:
    """Convert an iterable of Reaction objects to a pandas DataFrame.

    Parameters
    ----------
    reactions : Iterable[Reaction]
        Iterable of Reaction objects to convert.

    Returns
    -------
    pd.DataFrame
        DataFrame representation of the reactions.

    Raises
    ------
    IDNotSetError
        If any assembly ID in the reactions is not set.
    """
    rows = [
        _rename_id_key(ReactionRow.from_reaction(reaction).to_dict())
        for reaction in reactions
    ]

    return pd.DataFrame(rows)


def _rename_id_key(d: dict) -> dict:
    """Rename the 'id_' key in the dictionary to 'id'."""
    if 'id_' in d:
        d['id'] = d.pop('id_')
    return d



from dataclasses import dataclass
from typing import Mapping

import igraph as ig
from bidict import frozenbidict

from nasap_net.models import Assembly, BindingSite
from nasap_net.types import ID


@dataclass(frozen=True, init=False)
class GraphConversionResult:
    graph: ig.Graph
    core_mapping: frozenbidict[ID, int]
    binding_site_mapping: frozenbidict[BindingSite, int]

    def __init__(
            self,
            graph: ig.Graph,
            core_mapping: Mapping[ID, int],
            binding_site_mapping: Mapping[BindingSite, int],
    ) -> None:
        object.__setattr__(self, 'graph', graph)
        object.__setattr__(
            self, 'core_mapping', frozenbidict(core_mapping))
        object.__setattr__(
            self, 'binding_site_mapping',
            frozenbidict(binding_site_mapping))


def convert_assembly_to_igraph(assembly: Assembly) -> GraphConversionResult:
    g = ig.Graph()
    core_mapping = {}
    binding_site_mapping = {}

    for comp_id, comp in assembly.components.items():
        # Add the core node
        g.add_vertices(  # More efficient than add_vertex
            1,
            {
                'comp_id': [comp_id],
                'comp_kind': [comp.kind],
                'core_or_site': ['core'],
            }
        )
        core_mapping[comp_id] = g.vcount() - 1

        # Add the binding sites
        g.add_vertices(
            len(comp.site_ids),
            {
                'comp_id': [comp_id] * len(comp.site_ids),
                'comp_kind': [comp.kind] * len(comp.site_ids),
                'core_or_site': ['site'] * len(comp.site_ids),
                'site_id': list(comp.site_ids),
            }
        )
        start_id = g.vcount() - len(comp.site_ids)
        for i, site in enumerate(comp.get_binding_sites(comp_id)):
            binding_site_mapping[site] = start_id + i

        # Add the edges between core and sites
        g.add_edges(
            [(core_mapping[comp_id], binding_site_mapping[site])
                for site in comp.get_binding_sites(comp_id)]
        )

        # Add the auxiliary edges
        for aux in comp.aux_edges:
            site1, site2 = aux.get_binding_sites(comp_id)
            if aux.kind is None:
                g.add_edges(
                    [(
                        binding_site_mapping[site1],
                        binding_site_mapping[site2]
                    )]
                )
            else:
                g.add_edges(
                    [(
                        binding_site_mapping[site1],
                        binding_site_mapping[site2]
                    )],
                    {'aux_kind': [aux.kind]}
                )

    # Add the bonds
    for bond in assembly.bonds:
        site1, site2 = bond.sites
        g.add_edge(
            binding_site_mapping[site1],
            binding_site_mapping[site2]
        )

    return GraphConversionResult(
        graph=g,
        core_mapping=core_mapping,
        binding_site_mapping=binding_site_mapping
    )



import logging
from collections.abc import Iterable
from itertools import chain, product
from typing import Iterator, TypeVar

from nasap_net.helpers import validate_unique_assembly_ids
from nasap_net.models import Assembly, MLEKind, Reaction
from nasap_net.types import ID
from .explorer import InterReactionExplorer, IntraReactionExplorer
from .lib import ReactionOutOfScopeError, ReactionResolver

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

_T = TypeVar('_T', bound=ID)

def explore_reactions(
        assemblies: Iterable[Assembly],
        mle_kinds: Iterable[MLEKind],
        ) -> Iterator[Reaction]:
    logger.debug('Starting reaction exploration.')
    assemblies = list(assemblies)

    validate_unique_assembly_ids(assemblies)

    reaction_iters: list[Iterator[Reaction]] = []
    for mle_kind in mle_kinds:
        # Intra-molecular reactions
        for assem in assemblies:
            intra_explorer = IntraReactionExplorer(assem, mle_kind)
            reaction_iters.append(intra_explorer.explore())

        # Inter-molecular reactions
        for init_assem, entering_assem in product(assemblies, repeat=2):
            inter_explorer = InterReactionExplorer(
                init_assem, entering_assem, mle_kind)
            reaction_iters.append(inter_explorer.explore())

    resolver = ReactionResolver(assemblies)

    for i, reaction in enumerate(chain.from_iterable(reaction_iters)):
        try:
            resolved = resolver.resolve(reaction)
            logger.info('Reaction Found (%d): %s', i, resolved)
            yield resolved
        except ReactionOutOfScopeError:
            continue
    logger.debug('Reaction exploration completed.')



from nasap_net.models import Assembly
from .exceptions import NoIsomorphismFoundError
from .lib import color_vertices_and_edges, convert_assembly_to_igraph


def is_isomorphic(assem1: Assembly, assem2: Assembly) -> bool:
    g1 = convert_assembly_to_igraph(assem1).graph
    g2 = convert_assembly_to_igraph(assem2).graph

    try:
        colors = color_vertices_and_edges(g1, g2)
    except NoIsomorphismFoundError:
        return False

    return g1.isomorphic_vf2(
        g2,
        color1=colors.v_color1,
        color2=colors.v_color2,
        edge_color1=colors.e_color1,
        edge_color2=colors.e_color2,
    )



import pytest

from nasap_net.assembly_enumeration.lib.fragment_enumeration import \
    enumerate_fragments
from nasap_net.models import Assembly, AuxEdge, Bond, Component


@pytest.fixture
def M() -> Component:
    return Component(
        kind='M', sites=[0, 1, 2, 3],
        aux_edges=[AuxEdge(0, 1), AuxEdge(1, 2), AuxEdge(2, 3), AuxEdge(3, 0)])

@pytest.fixture
def L() -> Component:
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X() -> Component:
    return Component(kind='X', sites=[0])

@pytest.fixture
def MX2(M, X) -> Assembly:
    return Assembly(
        components={'M0': M, 'X0': X, 'X1': X},
        bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0)])


def test(MX2, M, X):
    frags = enumerate_fragments(MX2)
    assert frags == {
        Assembly(
            {'M0': M, 'X0': X, 'X1': X},
            [Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0)]
        ),
        Assembly({'M0': M, 'X0': X}, [Bond('M0', 0, 'X0', 0)]),
        Assembly({'M0': M, 'X1': X}, [Bond('M0', 1, 'X1', 0)]),
        Assembly({'M0': M}, []),
        Assembly({'X0': X}, []),
        Assembly({'X1': X}, []),
    }


def test_symmetry_operations(MX2, M, X):
    symmetry_operations = [
        {
            'M0': 'M0',
            'X0': 'X1',
            'X1': 'X0',
        }
    ]
    frags = enumerate_fragments(MX2, symmetry_operations)
    assert frags == {
        Assembly(
            {'M0': M, 'X0': X, 'X1': X},
            [Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0)]
        ),
        Assembly({'M0': M, 'X0': X}, [Bond('M0', 0, 'X0', 0)]),
        Assembly({'M0': M}, []),
        Assembly({'X0': X}, []),
    }


def test_M4L4(M4L4, M4L4_symmetry_operations):
    frags = enumerate_fragments(
        M4L4,
        list(M4L4_symmetry_operations.values())
    )
    assert len(frags) == 13


def test_M2L4(M2L4, M2L4_symmetry_operations):
    frags = enumerate_fragments(
        M2L4,
        list(M2L4_symmetry_operations.values())
    )
    assert len(frags) == 28


def test_M9L6(M9L6, M9L6_symmetry_operations):
    frags = enumerate_fragments(
        M9L6,
        list(M9L6_symmetry_operations.values())
    )
    assert len(frags) == 1480



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from types import MappingProxyType
from typing import Iterable

from nasap_net.models import Assembly
from nasap_net.types import ID
from .light_bond import LightBond


@dataclass(frozen=True)
class TemplateAssembly:
    assembly: Assembly

    @property
    def components(self) -> frozenset[ID]:
        return frozenset(self.assembly.components.keys())

    @property
    def bonds(self) -> frozenset[LightBond]:
        return frozenset(
            LightBond.from_bond(bond) for bond in self.assembly.bonds
        )

    def get_bonds_involving_components(
            self, components_ids: Iterable[ID]
    ) -> frozenset[LightBond]:
        """Return all bonds that involve at least one component ID from the
        specified set.
        """
        result_bonds: set[LightBond] = set()
        for comp_id in components_ids:
            bonds = self._component_id_to_bonds.get(comp_id, frozenset())
            result_bonds.update(bonds)
        return frozenset(result_bonds)

    @cached_property
    def _component_id_to_bonds(
            self
    ) -> MappingProxyType[ID, frozenset[LightBond]]:
        comp_id_to_bonds: dict[ID, set[LightBond]] = defaultdict(set)
        for bond in self.bonds:
            for comp_id in bond.component_ids:
                comp_id_to_bonds[comp_id].add(bond)
        return MappingProxyType({
            comp_id: frozenset(bonds)
            for comp_id, bonds in comp_id_to_bonds.items()
        })



from collections import UserDict
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any

from nasap_net.types import ID
from nasap_net.utils import resolve_chain_map


@dataclass
class SymmetryOperations(UserDict):
    data: dict[str, Mapping[Any, ID]] = field(default_factory=dict)

    def add_mapping(self, name: str, mapping: Mapping[Any, ID]) -> None:
        self.data[name] = mapping

    def add_cyclic_permutation(
            self,
            name: str,
            cyclic_permutation: Iterable[Sequence[ID]]
    ) -> None:
        self.add_mapping(name, cyclic_perms_to_map(cyclic_permutation))

    def add_product(
            self,
            name: str,
            mapping_names: Iterable[str]
    ) -> None:
        mappings = [
            self.data[mapping_name]
            for mapping_name in list(mapping_names)
        ]
        self.add_mapping(name, resolve_chain_map(*mappings))


def cyclic_perm_to_map(
        cyclic_permutation: Sequence[ID]) -> dict[ID, ID]:
    mapping: dict[ID, ID] = {}
    for i, source in enumerate(cyclic_permutation):
        # (i + 1) % N maps as follows:
        # 0 -> 1, 1 -> 2, ..., N-1 -> 0
        target = cyclic_permutation[(i + 1) % len(cyclic_permutation)]
        mapping[source] = target
    return mapping


def cyclic_perms_to_map(
        cyclic_permutations: Iterable[Sequence[ID]]) -> dict[ID, ID]:
    mapping = {}
    for perm in cyclic_permutations:
        mapping.update(cyclic_perm_to_map(perm))
    return mapping



from dataclasses import dataclass
from typing import Self

from nasap_net.exceptions import IDNotSetError
from nasap_net.models import Assembly, BindingSite
from nasap_net.types import ID
from nasap_net.utils.default import MISSING, Missing, default_if_missing


@dataclass(frozen=True, init=False)
class Reaction:
    init_assem: Assembly
    entering_assem: Assembly | None
    product_assem: Assembly
    leaving_assem: Assembly | None
    metal_bs: BindingSite
    leaving_bs: BindingSite
    entering_bs: BindingSite
    duplicate_count: int
    _id: ID | None

    def __init__(
            self,
            init_assem: Assembly,
            entering_assem: Assembly | None,
            product_assem: Assembly,
            leaving_assem: Assembly | None,
            metal_bs: BindingSite,
            leaving_bs: BindingSite,
            entering_bs: BindingSite,
            duplicate_count: int,
            id_: ID | None = None,
    ):
        if init_assem is None:
            raise TypeError("init_assem cannot be None")
        if product_assem is None:
            raise TypeError("product_assem cannot be None")
        if metal_bs is None:
            raise TypeError("metal_bs cannot be None")
        if leaving_bs is None:
            raise TypeError("leaving_bs cannot be None")
        if entering_bs is None:
            raise TypeError("entering_bs cannot be None")
        if duplicate_count is None:
            raise TypeError("duplicate_count cannot be None")
        if not duplicate_count > 0:
            raise ValueError("duplicate_count must be a positive integer")

        object.__setattr__(self, 'init_assem', init_assem)
        object.__setattr__(self, 'entering_assem', entering_assem)
        object.__setattr__(self, 'product_assem', product_assem)
        object.__setattr__(self, 'leaving_assem', leaving_assem)
        object.__setattr__(self, 'metal_bs', metal_bs)
        object.__setattr__(self, 'leaving_bs', leaving_bs)
        object.__setattr__(self, 'entering_bs', entering_bs)
        object.__setattr__(self, 'duplicate_count', duplicate_count)
        object.__setattr__(self, '_id', id_)

    def __str__(self):
        equation = self.equation_str
        dup = self.duplicate_count
        return f'{equation} (x{dup})'

    def __repr__(self):
        equation = self.equation_str
        if self._id is None:
            return f'<{self.__class__.__name__} {equation}>'
        return f'<{self.__class__.__name__} ID={self._id} {equation}>'

    @property
    def id_(self) -> ID:
        """Return the ID of the reaction."""
        if self._id is None:
            raise IDNotSetError("Reaction ID is not set.")
        return self._id

    @property
    def id_or_none(self) -> ID | None:
        """Return the ID of the reaction, or None if not set."""
        return self._id

    @property
    def equation_str(self) -> str:
        """Return a string representation of the reaction equation.

        If an assembly ID is not set, '??' is used in its place.
        """
        init = _assembly_to_id(self.init_assem)
        entering = _assembly_to_id(self.entering_assem)
        product = _assembly_to_id(self.product_assem)
        leaving = _assembly_to_id(self.leaving_assem)

        left = f'{init}' if entering is None else f'{init} + {entering}'
        right = f'{product}' if leaving is None else f'{product} + {leaving}'

        return f'{left} -> {right}'

    @property
    def init_assem_id(self) -> ID:
        """Return the ID of the initial assembly.

        Errors if the ID is not set.
        """
        return self.init_assem.id_

    @property
    def entering_assem_id(self) -> ID | None:
        """Return the ID of the entering assembly, or None if there is none.

        Errors if the ID is not set.
        """
        if self.entering_assem is None:
            return None
        return self.entering_assem.id_

    @property
    def product_assem_id(self) -> ID:
        """Return the ID of the product assembly.

        Errors if the ID is not set.
        """
        return self.product_assem.id_

    @property
    def leaving_assem_id(self) -> ID | None:
        """Return the ID of the leaving assembly, or None if there is none.

        Errors if the ID is not set.
        """
        if self.leaving_assem is None:
            return None
        return self.leaving_assem.id_

    def copy_with(
            self,
            *,
            init_assem: Assembly | Missing = MISSING,
            entering_assem: Assembly | None | Missing = MISSING,
            product_assem: Assembly | Missing = MISSING,
            leaving_assem: Assembly | None | Missing = MISSING,
            metal_bs: BindingSite | Missing = MISSING,
            leaving_bs: BindingSite | Missing = MISSING,
            entering_bs: BindingSite | Missing = MISSING,
            duplicate_count: int | Missing = MISSING,
            id_: ID | Missing = MISSING,
            ) -> Self:
        """Return a copy of the reaction with optional modifications.

        - Fields not provided will default to the current values,
            except for the ID, which will be set to None if not provided.
        - Fields explicitly set to None will overwrite with None.
            (only applies to fields that can be None)

        If you want to copy the current ID, specify it explicitly,
        e.g., `copied = reaction.copy_with(id_=reaction.id_or_none)`.
        """
        return self.__class__(
            init_assem=default_if_missing(init_assem, self.init_assem),
            entering_assem=default_if_missing(entering_assem, self.entering_assem),
            product_assem=default_if_missing(product_assem, self.product_assem),
            leaving_assem=default_if_missing(leaving_assem, self.leaving_assem),
            metal_bs=default_if_missing(metal_bs, self.metal_bs),
            leaving_bs=default_if_missing(leaving_bs, self.leaving_bs),
            entering_bs=default_if_missing(entering_bs, self.entering_bs),
            duplicate_count=default_if_missing(duplicate_count, self.duplicate_count),
            id_=default_if_missing(id_, None),
        )


def _assembly_to_id(assembly: Assembly | None) -> ID | None:
    """Return the ID of the assembly, or '??' if not set, or None if assembly
    is None.
    """
    if assembly is None:
        return None
    if assembly.id_or_none is None:
        # ID not set
        return '??'
    return assembly.id_or_none



from typing import Any, NewType, TypeVar, overload


Missing = NewType("Missing", object)
MISSING = Missing(object())

T = TypeVar('T')

@overload
def default_if_missing(value: Missing, default: T) -> T: ...
@overload
def default_if_missing(value: T, default: Any) -> T: ...
def default_if_missing(value, default):
    """Return the value if it is not MISSING, otherwise return the default.

    Note that None is considered a valid value and will not trigger the default.
    """
    return default if value is MISSING else value



"""Public helpers for the nasap_net package."""

from .assembly_indexing import assign_composition_formula_ids
from .composition_formula import generate_composition_formula
from .unique_assembly_id_validation import validate_unique_assembly_ids



from collections.abc import Hashable, Iterable
from typing import Protocol, Self, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec('P')


class HasCopyWith(Protocol[P]):
    @property
    def id_(self) -> Hashable:
        ...

    # TODO: Use more specific signature when PyCharm fixes the issue PY-70838
    # copy_with(self, *, id_: Any) -> Self:
    def copy_with(self, **kwargs: P.kwargs) -> Self:
        ...


_T = TypeVar('_T', bound=HasCopyWith)

def deduplicate_ids(items: Iterable[_T]) -> list[_T]:
    """Deduplicate IDs of items by appending suffixes.

    If multiple items have the same ID, this function appends suffixes
    like '_2', '_3', etc.

    - The first occurrence of an ID remains unchanged.
    - The second occurrence gets '_2' appended.
    - The third occurrence gets '_3' appended, and so on.

    Parameters
    ----------
    items : Iterable[_T]
        An iterable of items with IDs.

    Returns
    -------
    list[_T]
        A list of items with deduplicated IDs.

    Notes
    -----
    This function assumes that each item has an 'id_' attribute and a
    'copy_with' method that allows creating a copy of the item with
    modified attributes.
    The `copy_with` method should accept keyword arguments of `id_`.
    """
    seen_ids: set[Hashable] = set()
    deduplicated_items: list[_T] = []

    for item in items:
        if item.id_ not in seen_ids:
            seen_ids.add(item.id_)
            deduplicated_items.append(item)
            continue

        suffix = 2
        new_id = f'{item.id_}_{suffix}'
        while new_id in seen_ids:
            suffix += 1
            new_id = f'{item.id_}_{suffix}'
        seen_ids.add(new_id)
        deduplicated_items.append(item.copy_with(id_=new_id))
    return deduplicated_items



from dataclasses import dataclass
from functools import total_ordering

from nasap_net.types import ID, SupportsDunderLt


@total_ordering
@dataclass(frozen=True)
class BindingSite(SupportsDunderLt):
    """A specific binding site on a specific component."""
    component_id: ID
    site_id: ID

    def __lt__(self, other):
        if not isinstance(other, BindingSite):
            return NotImplemented
        self_values = (self.component_id, self.site_id)
        other_values = (other.component_id, other.site_id)
        return self_values < other_values



from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from functools import cached_property, total_ordering
from types import MappingProxyType
from typing import Any, Self

from frozendict import frozendict

from nasap_net.exceptions import IDNotSetError, NasapNetError
from nasap_net.types import ID
from nasap_net.utils import construct_repr
from nasap_net.utils.default import MISSING, Missing, default_if_missing
from .binding_site import BindingSite
from .bond import Bond
from .component import Component


@dataclass
class InvalidBondError(NasapNetError):
    bond: Bond
    detail: str | None = None

    def __str__(self) -> str:
        base_msg = f'Invalid bond: {self.bond}'
        if self.detail:
            return f'{base_msg} - {self.detail}'
        return base_msg


@dataclass
class InconsistentComponentError(NasapNetError):
    """Raised when there are inconsistent definitions for a component kind,
    i.e., the same kind name corresponds to different component structures.
    """
    component_kind: str
    component1: Component
    component2: Component

    def __str__(self) -> str:
        return (
            f'Inconsistent definitions for component kind '
            f'"{self.component_kind}": '
            f'Component 1: {self.component1}, Component 2: {self.component2}.'
        )


@dataclass
class ParallelBondError(NasapNetError):
    bond1: Bond
    bond2: Bond

    def __str__(self) -> str:
        return (
            f'Multiple bonds between the same pair of components '
            f'are not supported: '
            f'Bond 1: {self.bond1}, Bond 2: {self.bond2}.'
        )


@dataclass
class BondNotFoundError(NasapNetError):
    comp_id1: ID
    comp_id2: ID

    def __str__(self) -> str:
        return (
            f'No bond found between components '
            f'{self.comp_id1} and {self.comp_id2}.'
        )


@total_ordering
@dataclass(frozen=True, init=False)
class Assembly:
    """An assembly of components connected by bonds.

    Parameters
    ----------
    components : Mapping[C, nasap_net.models.component.Component[S]]
        A mapping from component IDs to their corresponding components.
    bonds : Iterable[Bond[C, S]]
        An iterable of bonds connecting the components.

    Raises
    ------
    - InconsistentComponentError
        If components with the same kind have different structures.
    - InvalidBondError
        - If any bond references a non-existent component or site.
        - If a component bonds to itself.
        - If a site is used more than once.
        - If the assembly is not connected.
    - ParallelBondError
        If there are multiple bonds between the same pair of components.

    Warnings
    --------
    - The assembly does not enforce connectivity; it is the user's
      responsibility to ensure that the assembly is connected as needed.

    Notes
    -----
    Currently, the assembly does not support parallel bonds (multiple bonds
    between the same pair of components), e.g., chelate complexes.
    """
    _components: frozendict[ID, Component]
    bonds: frozenset[Bond]
    _id: ID | None

    def __init__(
            self, components: Mapping[ID, Component],
            bonds: Iterable[Bond],
            *,
            id_: ID | None = None
            ):
        if components is None:
            raise TypeError("components cannot be None")
        if bonds is None:
            raise TypeError("bonds cannot be None")

        object.__setattr__(self, '_components', frozendict(components))
        object.__setattr__(self, 'bonds', frozenset(bonds))
        object.__setattr__(self, '_id', id_)
        self._validate()

    def __lt__(self, other):
        if not isinstance(other, Assembly):
            return NotImplemented
        # 1. number of components
        # 2. number of bonds
        # 3. component IDs (sorted)
        # 4. bonds (sorted)
        if len(self._components) != len(other._components):
            return len(self._components) < len(other._components)
        if len(self.bonds) != len(other.bonds):
            return len(self.bonds) < len(other.bonds)
        self_comp_ids = sorted(self._components.keys())
        other_comp_ids = sorted(other._components.keys())
        if self_comp_ids != other_comp_ids:
            return self_comp_ids < other_comp_ids
        self_bonds = sorted(self.bonds)
        other_bonds = sorted(other.bonds)
        return self_bonds < other_bonds

    def __repr__(self):
        fields: dict[str, Any] = {}
        if self._id is not None:
            fields['id_'] = self._id
        fields['components'] = dict(sorted(self.component_id_to_kind.items()))
        fields['bonds'] = [bond.to_tuple() for bond in sorted(self.bonds)]
        return construct_repr(self.__class__, fields)

    @property
    def id_(self) -> ID:
        """Return the ID of the assembly."""
        if self._id is None:
            raise IDNotSetError("Assembly ID is not set.")
        return self._id

    @property
    def id_or_none(self) -> ID | None:
        """Return the ID of the assembly, or None if not set."""
        return self._id

    @property
    def components(self) -> Mapping[ID, Component]:
        """Return the components in the assembly as an immutable mapping."""
        return MappingProxyType(self._components)

    @cached_property
    def component_id_to_kind(self) -> Mapping[ID, str]:
        """Return a mapping from component IDs to their kinds."""
        return MappingProxyType({
            comp_id: comp.kind for comp_id, comp
            in self._components.items()
        })

    def get_component_kind_of_site(self, site: BindingSite) -> str:
        """Return the component kind of the given binding site."""
        return self._components[site.component_id].kind

    def find_sites(
            self, *, has_bond: bool | None = None,
            component_kind: str | None = None
            ) -> frozenset[BindingSite]:
        """Return binding sites based on their bond status.
        """
        if has_bond is None and component_kind is None:
            return self._all_sites

        sites = set()
        for site in self._all_sites:
            if has_bond is not None:
                if has_bond != self.has_bond(site):
                    continue
            if component_kind is not None:
                if self.get_component_kind_of_site(site) != component_kind:
                    continue
            sites.add(site)
        return frozenset(sites)

    def has_bond(self, site: BindingSite) -> bool:
        """Check if a binding site has a bond."""
        return site in self._sites_with_bond

    def has_bond_between_components(self, comp_id1: ID, comp_id2: ID) -> bool:
        """Check if there is a bond between two components."""
        return frozenset({comp_id1, comp_id2}) in self._component_connection

    def get_bond_by_comp_ids(self, comp_id1: ID, comp_id2: ID) -> Bond:
        """Return the bond between two components.

        Parameters
        ----------
        comp_id1 : ID
            The ID of the first component.
        comp_id2 : ID
            The ID of the second component.

        Returns
        -------
        Bond
            The bond between the two components.

        Raises
        ------
        BondNotFoundError
            If there is no bond between the two components.

        Notes
        -----
        This method assumes that there is no parallel bonds between the
        same pair of components.
        """
        for bond in self.bonds:
            if bond.component_ids == frozenset({comp_id1, comp_id2}):
                return bond
        raise BondNotFoundError(comp_id1=comp_id1, comp_id2=comp_id2)

    def add_bond(self, site1: BindingSite, site2: BindingSite):
        """Return a new assembly with an additional bond."""
        new_bond = Bond.from_sites(site1, site2)
        new_bonds = set(self.bonds)
        new_bonds.add(new_bond)
        return self.copy_with(bonds=new_bonds)

    def remove_bond(self, site1: BindingSite, site2: BindingSite):
        """Return a new assembly with a bond removed."""
        bond_to_remove = Bond.from_sites(site1, site2)
        new_bonds = set(self.bonds)
        new_bonds.remove(bond_to_remove)
        return self.copy_with(bonds=new_bonds)

    def copy_with(
            self,
            *,
            components: Mapping[ID, Component] | Missing = MISSING,
            bonds: Iterable[Bond] | Missing = MISSING,
            id_: ID | Missing = MISSING,
            ) -> Self:
        """Return a copy of the assembly with optional modifications.

        Fields not provided will default to the current values, except for the
        ID, which will be set to None if not provided.

        If you want to copy the current ID, specify it explicitly,
        e.g., `copied = assembly.copy_with(id_=assembly.id_or_none)`.
        """
        return self.__class__(
            components=default_if_missing(components, self._components),
            bonds=default_if_missing(bonds, self.bonds),
            id_=default_if_missing(id_, None),
        )

    @cached_property
    def _all_sites(self) -> frozenset[BindingSite]:
        """Return all binding sites in the assembly."""
        sites = set()
        for comp_id, component in self._components.items():
            for site in component.site_ids:
                sites.add(BindingSite(component_id=comp_id, site_id=site))
        return frozenset(sites)

    @cached_property
    def _sites_with_bond(self) -> frozenset[BindingSite]:
        """Return a set of binding sites that have bonds."""
        site_with_bonds = set()
        for bond in self.bonds:
            for site in bond.sites:
                site_with_bonds.add(site)
        return frozenset(site_with_bonds)

    @cached_property
    def _component_connection(self) -> frozenset[frozenset[ID]]:
        connections = set()
        for bond in self.bonds:
            comp_ids = frozenset(bond.component_ids)
            connections.add(comp_ids)
        return frozenset(connections)

    def _get_component_of_site(self, site: BindingSite) -> Component:
        """Return the component corresponding to the given binding site."""
        return self._components[site.component_id]

    def _validate(self):
        self._validate_components()
        self._validate_bonds()
        self._validate_parallel_bonds()

    def _validate_components(self):
        # Components with the same kind should have the same structure.
        comp_kind_to_obj: dict[str, Component] = {}
        for comp in self._components.values():
            if comp.kind in comp_kind_to_obj:
                if comp != comp_kind_to_obj[comp.kind]:
                    raise InconsistentComponentError(
                        component_kind=comp.kind,
                        component1=comp,
                        component2=comp_kind_to_obj[comp.kind],
                    )
            else:
                comp_kind_to_obj[comp.kind] = comp

    def _validate_bonds(self):
        component_keys = set(self._components.keys())
        used_sites = set()
        for bond in self.bonds:
            # Validate that the components exist
            for comp_id in bond.component_ids:
                if comp_id not in component_keys:
                    raise InvalidBondError(
                        bond=bond,
                        detail=f"Component {comp_id} not found in assembly.")

            # Validate that the sites exist in the respective components
            for site in bond.sites:
                component = self._components[site.component_id]
                if site.site_id not in component.site_ids:
                    raise InvalidBondError(
                        bond=bond,
                        detail=(
                            f"Site {site.site_id} not found in component "
                            f"{site.component_id}."))

                # Validate that the site is not already used
                if site in used_sites:
                    raise InvalidBondError(
                        bond=bond,
                        detail=f"Site {site} is already used in another bond.")
                used_sites.add(site)

    def _validate_parallel_bonds(self):
        # Currently, parallel bonds (multiple bonds between the same pair
        # of components) are not supported.
        bonded_pairs: dict[frozenset[ID], Bond] = {}
        for bond in self.bonds:
            comp_pair = frozenset(bond.component_ids)
            if comp_pair in bonded_pairs:
                raise ParallelBondError(
                    bond1=bonded_pairs[comp_pair], bond2=bond
                )
            bonded_pairs[comp_pair] = bond



from collections.abc import Mapping
from typing import Any, TypeVar

from nasap_net.exceptions import NasapNetError
from nasap_net.types import ID
from ..models import Fragment, LightBond

_T = TypeVar('_T', bound=ID)


def validate_symmetry_operation(
        fragment: Fragment,
        symmetry_operation: Mapping[_T, _T]
) -> None:
    if set(symmetry_operation.keys()) != set(fragment.components):
        raise InvalidSymmetryOperationError()
    if set(symmetry_operation.values()) != set(fragment.components):
        raise InvalidSymmetryOperationError()


def apply_symmetry_operation(
        fragment: Fragment,
        symmetry_operation: Mapping[Any, ID]
) -> Fragment:
    components = {
        symmetry_operation[comp_id] for comp_id in fragment.components
    }
    bonds = [
        LightBond(
            *(symmetry_operation[comp_id] for comp_id in bond.component_ids)
        )
        for bond in fragment.bonds
    ]
    return fragment.copy_with(
        components=components,
        bonds=bonds,
    )


class InvalidSymmetryOperationError(NasapNetError):
    pass



from collections.abc import Hashable, Iterable, Mapping
from typing import Any

from nasap_net.types import ID
from .key import get_key
from .symmetry_operation import apply_symmetry_operation
from ..models import Fragment


def is_new(
        target: Fragment,
        found: Mapping[Hashable, set[Fragment]],
        symmetry_operations: Iterable[Mapping[Any, ID]] | None = None
) -> bool:
    key = get_key(target)
    if key not in found:
        return True
    candidates = found[key]
    if target in candidates:
        return False
    if symmetry_operations is None:
        return True
    for sym_op in symmetry_operations:
        transformed_fragment = apply_symmetry_operation(target, sym_op)
        if transformed_fragment in candidates:
            return False
    return True



from collections.abc import Hashable
from functools import lru_cache

from ..models import Fragment


@lru_cache
def get_key(fragment: Fragment) -> Hashable:
    return len(fragment.components), len(fragment.bonds)



from collections.abc import Iterable

from nasap_net.types import ID
from ..models import Fragment


def create_sub_fragment(
        fragment: Fragment, comp_ids: Iterable[ID]
) -> Fragment:
    """Create a sub-assembly containing only the specified components."""
    new_components = {
        comp_id for comp_id in fragment.components
        if comp_id in comp_ids
    }
    new_bonds = [
        bond for bond in fragment.bonds
        if bond.component_ids <= set(comp_ids)
    ]
    return fragment.copy_with(components=new_components, bonds=new_bonds)



from typing import Sequence

from nasap_net.models import Assembly
from nasap_net.utils import deduplicate_ids
from .composition_formula import generate_composition_formula


def assign_composition_formula_ids(
        assemblies: Sequence[Assembly],
        *,
        order: Sequence[str] | None = None,
) -> list[Assembly]:
    """Assign composition formula strings as IDs to the given assemblies."""
    assigned_assemblies = [
        assembly.copy_with(
            id_=generate_composition_formula(assembly, order=order)
        )
        for assembly in assemblies
    ]
    return deduplicate_ids(assigned_assemblies)  # type: ignore



from nasap_net.helpers import generate_composition_formula
from nasap_net.models import Assembly, Component


def test_composition_formula():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])

    MX2 = Assembly(
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[],
    )

    assert generate_composition_formula(MX2) == 'MX2'


def test_order():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])

    MX2 = Assembly(
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[],
    )

    assert generate_composition_formula(MX2) == 'MX2'
    assert generate_composition_formula(MX2, order=['X', 'M']) == 'X2M'


def test_unspecified_components_are_sorted_alphabetically():
    A = Component(kind='A', sites=[0])
    B = Component(kind='B', sites=[0])
    C = Component(kind='C', sites=[0])
    D = Component(kind='D', sites=[0])

    ABCD = Assembly({'A0': A, 'B0': B, 'C0': C, 'D0': D}, bonds=[])

    assert generate_composition_formula(ABCD, order=['C', 'A']) == 'CA' + 'BD'



import pytest

from nasap_net.io import load_reactions, save_reactions
from nasap_net.models import Assembly, BindingSite, Bond, Component, Reaction


@pytest.fixture
def M():
    return Component(kind='M', sites=[0, 1])

@pytest.fixture
def L():
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X():
    return Component(kind='X', sites=[0])

@pytest.fixture
def MX2(M, X):
    return Assembly(
        id_='MX2',
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]
    )

@pytest.fixture
def free_L(L):
    return Assembly(id_='free_L', components={'L0': L}, bonds=[])

@pytest.fixture
def MLX(M, L, X):
    return Assembly(
        id_='MLX',
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[],
    )

@pytest.fixture
def free_X(X):
    return Assembly(id_='free_X', components={'X0': X}, bonds=[])

@pytest.fixture
def MX2_and_free_L(M, L, X, MX2, free_L, MLX, free_X):
    return Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 0),
        duplicate_count=4,
        id_='R1',
    )


def test_basic(tmp_path, MX2_and_free_L, MX2, free_L, MLX, free_X):
    output_file = tmp_path / 'reactions.csv'
    save_reactions([MX2_and_free_L], output_file)

    loaded_reactions = load_reactions(
        output_file,
        assemblies=[MX2, free_L, MLX, free_X],
        site_id_type='int'
    )

    assert len(loaded_reactions) == 1
    assert loaded_reactions[0] == MX2_and_free_L


def test_nan_handling(tmp_path):
    """Check if NaN values are handled correctly.

    - no entering_assem
    - no leaving_assem
    - no reaction ID
    """
    reaction_with_nan = Reaction(
        init_assem=Assembly(id_='A1', components={}, bonds=[]),
        entering_assem=None,
        product_assem=Assembly(id_='A2', components={}, bonds=[]),
        leaving_assem=None,
        metal_bs=BindingSite('M0', '0'),
        leaving_bs=BindingSite('X0', '0'),
        entering_bs=BindingSite('L0', '0'),
        duplicate_count=1,
        id_=None,
    )

    output_file = tmp_path / 'reactions.csv'
    save_reactions([reaction_with_nan], output_file)

    loaded_reactions = load_reactions(
        output_file,
        assemblies=[
            Assembly(id_='A1', components={}, bonds=[]),
            Assembly(id_='A2', components={}, bonds=[]),
        ],
    )

    assert len(loaded_reactions) == 1
    assert loaded_reactions[0] == reaction_with_nan


def test_types(tmp_path):
    int_comp = Component(kind='kind', sites=[100])
    int_assem = Assembly(id_=300, components={200: int_comp}, bonds=[])
    int_reaction = Reaction(
        init_assem=int_assem,
        entering_assem=None,
        product_assem=int_assem,
        leaving_assem=None,
        metal_bs=BindingSite(200, 100),
        leaving_bs=BindingSite(200, 100),
        entering_bs=BindingSite(200, 100),
        duplicate_count=1,
        id_=400,
    )
    output_file = tmp_path / 'reactions.csv'
    save_reactions([int_reaction], output_file)

    loaded_all_int = load_reactions(
        output_file,
        assemblies=[int_assem],
        assembly_id_type='int',
        component_id_type='int',
        site_id_type='int',
        reaction_id_type='int',
    )
    assert len(loaded_all_int) == 1
    assert loaded_all_int[0] == int_reaction

    str_comp = Component(kind='kind', sites=['100'])
    str_assem = Assembly(id_='300', components={'200': str_comp}, bonds=[])
    str_reaction = Reaction(
        init_assem=str_assem,
        entering_assem=None,
        product_assem=str_assem,
        leaving_assem=None,
        metal_bs=BindingSite('200', '100'),
        leaving_bs=BindingSite('200', '100'),
        entering_bs=BindingSite('200', '100'),
        duplicate_count=1,
        id_='400',
    )
    loaded_all_str = load_reactions(
        output_file,
        assemblies=[str_assem],
        assembly_id_type='str',
        component_id_type='str',
        site_id_type='str',
        reaction_id_type='str',
    )
    assert len(loaded_all_str) == 1
    assert loaded_all_str[0] == str_reaction



import os
from collections.abc import Iterable
from pathlib import Path

from nasap_net.io.assemblies.dump import dump
from nasap_net.models import Assembly


def save_assemblies(
        assemblies: Iterable[Assembly],
        file_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        verbose: bool = True,
        ) -> None:
    """Dump assemblies and components into a YAML file.

    Parameters
    ----------
    assemblies : Iterable[Assembly]
        Assemblies to dump.
    file_path : os.PathLike | str
        Path to the YAML file to write.
    overwrite : bool, optional
        If True, overwrite the file if it already exists.
        If False, raise an error if the file already exists.
        Default is False.
    verbose : bool, optional
        If True, print a message indicating the file path written to.
        Default is True.
    """
    file_path = Path(file_path)
    if file_path.exists() and not overwrite:
        raise FileExistsError(
            f'File "{str(file_path)}" already exists. '
            'Use `overwrite=True` to overwrite it.'
        )

    file_path.parent.mkdir(parents=True, exist_ok=True)
    yaml_str = dump(assemblies)
    file_path.write_text(yaml_str, encoding="utf-8")
    if verbose:
        print(f'Saved! --> "{str(file_path)}"')



from .assembly import Assembly
from .aux_edge import AuxEdge
from .binding_site import BindingSite
from .bond import Bond
from .component import Component
from .mle import DuplicationNotSetError, MLE, MLEKind
from .reaction import Reaction



from collections.abc import Iterable
from dataclasses import dataclass

from nasap_net.exceptions import NasapNetError
from nasap_net.models import Assembly, Component


@dataclass(frozen=True)
class FoundComponent:
    component: Component
    source_assembly: Assembly


def check_component_consistency(assemblies: Iterable[Assembly]) -> None:
    """Check for consistent definitions of component kinds across assemblies.

    Components with the same kind name must have identical structures,
    including site IDs.

    The function raises an InconsistentComponentBetweenAssembliesError
    if it finds any inconsistencies.

    Parameters
    ----------
    assemblies : Iterable[Assembly]
        An iterable of Assembly instances to check for component consistency.

    Raises
    ------
    InconsistentComponentBetweenAssembliesError
        If there are inconsistent definitions for a component kind,
        i.e., the same kind name corresponds to different component structures.
    """
    found_components: dict[str, FoundComponent] = {}
    for assembly in assemblies:
        for comp in assembly.components.values():
            if comp.kind in found_components:
                if comp != found_components[comp.kind].component:
                    raise InconsistentComponentBetweenAssembliesError(
                        component_kind=comp.kind,
                        assembly1=found_components[comp.kind].source_assembly,
                        assembly2=assembly,
                    )
            else:
                found_components[comp.kind] = FoundComponent(
                    component=comp,
                    source_assembly=assembly
                )


@dataclass
class InconsistentComponentBetweenAssembliesError(NasapNetError):
    """Raised when there are inconsistent definitions for a component kind
    between different assemblies.
    """
    component_kind: str
    assembly1: Assembly
    assembly2: Assembly

    def __str__(self) -> str:
        return (
            f'Inconsistent definitions for component kind '
            f'"{self.component_kind}" between assemblies: '
            f'Assembly 1: {self.assembly1}, Assembly 2: {self.assembly2}.'
        )



import pandas as pd
import pytest

from nasap_net.io import load_reactions
from nasap_net.models import Assembly, BindingSite, Bond, Component, Reaction


@pytest.fixture
def M():
    return Component(kind='M', sites=[0, 1])

@pytest.fixture
def L():
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X():
    return Component(kind='X', sites=[0])

@pytest.fixture
def MX2(M, X):
    return Assembly(
        id_='MX2',
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]
    )

@pytest.fixture
def free_L(L):
    return Assembly(id_='free_L', components={'L0': L}, bonds=[])

@pytest.fixture
def MLX(M, L, X):
    return Assembly(
        id_='MLX',
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[],
    )

@pytest.fixture
def free_X(X):
    return Assembly(id_='free_X', components={'X0': X}, bonds=[])


def test_basic(tmp_path, MX2, free_L, MLX, free_X):
    df = pd.DataFrame([{
        'init_assem_id': 'MX2',
        'entering_assem_id': 'free_L',
        'product_assem_id': 'MLX',
        'leaving_assem_id': 'free_X',
        'metal_bs_component': 'M0',
        'metal_bs_site': 0,
        'leaving_bs_component': 'X0',
        'leaving_bs_site': 0,
        'entering_bs_component': 'L0',
        'entering_bs_site': 0,
        'duplicate_count': 4,
    }])

    reaction_file = tmp_path / 'reactions.csv'
    df.to_csv(reaction_file, index=False)

    imported_reactions = load_reactions(
        reaction_file,
        assemblies=[MX2, free_L, MLX, free_X],
        assembly_id_type='str',
        component_id_type='str',
        site_id_type='str',
    )

    assert len(imported_reactions) == 1
    assert imported_reactions[0] == Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', '0'),
        leaving_bs=BindingSite('X0', '0'),
        entering_bs=BindingSite('L0', '0'),
        duplicate_count=4,
    )


def test_index_column(tmp_path, MX2, free_L, MLX, free_X):
    df = pd.DataFrame([{
        'init_assem_id': 'MX2',
        'entering_assem_id': 'free_L',
        'product_assem_id': 'MLX',
        'leaving_assem_id': 'free_X',
        'metal_bs_component': 'M0',
        'metal_bs_site': 0,
        'leaving_bs_component': 'X0',
        'leaving_bs_site': 0,
        'entering_bs_component': 'L0',
        'entering_bs_site': 0,
        'duplicate_count': 4,
    }])
    reaction_file = tmp_path / 'reactions_with_index.csv'
    df.to_csv(reaction_file, index_label='index_column')

    imported_reactions = load_reactions(
        reaction_file,
        assemblies=[MX2, free_L, MLX, free_X],
        assembly_id_type='str',
        component_id_type='str',
        site_id_type='str',
        has_index_column=True,
    )

    assert len(imported_reactions) == 1
    assert imported_reactions[0] == Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', '0'),
        leaving_bs=BindingSite('X0', '0'),
        entering_bs=BindingSite('L0', '0'),
        duplicate_count=4,
    )


def test_with_id_column(tmp_path, MX2, free_L, MLX, free_X):
    df = pd.DataFrame([{
        'init_assem_id': 'MX2',
        'entering_assem_id': 'free_L',
        'product_assem_id': 'MLX',
        'leaving_assem_id': 'free_X',
        'metal_bs_component': 'M0',
        'metal_bs_site': 0,
        'leaving_bs_component': 'X0',
        'leaving_bs_site': 0,
        'entering_bs_component': 'L0',
        'entering_bs_site': 0,
        'duplicate_count': 4,
        'id': 'R1',
    }])
    reaction_file = tmp_path / 'reactions_with_id.csv'
    df.to_csv(reaction_file, index=False)

    imported_reactions = load_reactions(
        reaction_file,
        assemblies=[MX2, free_L, MLX, free_X],
        assembly_id_type='str',
        component_id_type='str',
        site_id_type='str',
        reaction_id_type='str',
    )

    assert len(imported_reactions) == 1
    assert imported_reactions[0] == Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', '0'),
        leaving_bs=BindingSite('X0', '0'),
        entering_bs=BindingSite('L0', '0'),
        duplicate_count=4,
        id_='R1',
    )



from collections import defaultdict
from collections.abc import Hashable, Iterable, Iterator, Mapping
from typing import Any

from nasap_net.types import ID
from .is_new import is_new
from .key import get_key
from ..models import Fragment


def extract_unique_fragments(
        fragments: Iterable[Fragment],
        symmetry_operations: Iterable[Mapping[Any, ID]] | None = None
) -> Iterator[Fragment]:
    unique_fragments: dict[Hashable, set[Fragment]] = defaultdict(set)
    for fragment in fragments:
        if is_new(fragment, unique_fragments, symmetry_operations):
            unique_fragments[get_key(fragment)].add(fragment)
            yield fragment



from dataclasses import FrozenInstanceError

import pytest

from nasap_net.models import Component


def test_component():
    M = Component(kind='M', sites=['a', 'b'])
    assert M.kind == 'M'
    assert M.site_ids == frozenset({'a', 'b'})


def test_component_immutability():
    M = Component(kind='M', sites=['a', 'b'])
    with pytest.raises(FrozenInstanceError):
        M.kind = 'M2'  # type: ignore[attr-defined, misc]
    with pytest.raises(FrozenInstanceError):
        M.site_ids = frozenset({'c'})  # type: ignore[attr-defined, misc]


def test___repr__():
    M = Component(kind='M', sites=[0, 1])
    assert repr(M) == "<Component kind='M', site_ids=[0, 1]>"


def test___repr___with_aux_edges():
    from nasap_net.models import AuxEdge

    M = Component(
        kind='M',
        sites=[0, 1, 2],
        aux_edges=[AuxEdge(0, 1, 'cis'), AuxEdge(1, 2)],
    )
    assert repr(M) == (
        "<Component kind='M', site_ids=[0, 1, 2], "
        "aux_edges=[(0, 1, 'cis'), (1, 2)]>"
    )



from collections.abc import Mapping
from typing import TypeVar

from nasap_net.io.assemblies.semi_light_assembly import SemiLightAssembly
from nasap_net.models import Assembly, Component
from nasap_net.types import ID

_T = TypeVar('_T', bound=ID)

def convert_semi_light_assembly_to_rich_one(
        semi_light_assembly: SemiLightAssembly,
        components: Mapping[str, Component],
) -> Assembly:
    return Assembly(
        components={
            comp_id: components[comp_kind]
            for comp_id, comp_kind in semi_light_assembly.components.items()
        },
        bonds=semi_light_assembly.bonds,
        id_=semi_light_assembly.id_or_none,
    )


def convert_semi_light_assemblies_to_rich_ones(
        semi_light_assemblies: Mapping[_T, SemiLightAssembly],
        components: Mapping[str, Component],
) -> dict[ID, Assembly]:
    return {
        assembly_id: convert_semi_light_assembly_to_rich_one(
            semi_light_assembly, components
        )
        for assembly_id, semi_light_assembly in semi_light_assemblies.items()
    }



from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import TypeVar

from nasap_net.models import Assembly, Component
from nasap_net.models.component_consistency_check import \
    check_component_consistency
from nasap_net.types import ID
from .semi_light_assembly import SemiLightAssembly


@dataclass(frozen=True)
class ConversionResult:
    semi_light_assemblies: Mapping[ID, SemiLightAssembly]
    components: Mapping[str, Component]


_T = TypeVar('_T', bound=ID)

def convert_assemblies_to_semi_light_ones(
        assemblies: Mapping[_T, Assembly],
) -> ConversionResult:
    semi_light_assemblies = _assemblies_to_semi_light_assemblies(assemblies)

    check_component_consistency(assemblies.values())
    components = _extract_components(assemblies.values())

    return ConversionResult(
        semi_light_assemblies=semi_light_assemblies,
        components=components,
    )


def _extract_components(
        assemblies: Iterable[Assembly],
) -> dict[str, Component]:
    components: dict[str, Component] = {}
    for assembly in assemblies:
        for comp in assembly.components.values():
            if comp.kind in components:
                assert comp == components[comp.kind]
            else:
                components[comp.kind] = comp
    return components


def _assemblies_to_semi_light_assemblies(
        assemblies: Mapping[_T, Assembly],
        ) -> Mapping[ID, SemiLightAssembly]:
    return {
        assembly_id: SemiLightAssembly.from_assembly(assembly)
        for assembly_id, assembly in assemblies.items()
    }



import pytest

from nasap_net.models import BindingSite, Bond


def test_bond():
    bond = Bond(comp_id1="L1", site1="a", comp_id2="M1", site2="b")
    assert bond.sites == frozenset({
        BindingSite(component_id="L1", site_id="a"),
        BindingSite(component_id="M1", site_id="b")
    })


def test_bond_same_site_error():
    with pytest.raises(ValueError):
        Bond(comp_id1="M1", site1="a", comp_id2="M1", site2="b")
    # Should not raise
    Bond(comp_id1="M1", site1="a", comp_id2="M2", site2="a")



import pytest

from nasap_net.exceptions import DuplicateIDError, IDNotSetError
from nasap_net.helpers import validate_unique_assembly_ids
from nasap_net.models import Assembly


def test_valid():
    assemblies = [
        Assembly(id_='A1', components={}, bonds=[]),
        Assembly(id_='A2', components={}, bonds=[]),
        Assembly(id_='A3', components={}, bonds=[]),
    ]
    validate_unique_assembly_ids(assemblies)


def test_missing_id():
    assemblies = [
        Assembly(id_='A1', components={}, bonds=[]),
        Assembly(components={}, bonds=[]),  # Missing ID
        Assembly(id_='A3', components={}, bonds=[]),
    ]
    with pytest.raises(IDNotSetError):
        validate_unique_assembly_ids(assemblies)


def test_duplicate_id():
    assemblies = [
        Assembly(id_='A1', components={}, bonds=[]),
        Assembly(id_='A2', components={}, bonds=[]),
        Assembly(id_='A1', components={}, bonds=[]),  # Duplicate ID
    ]
    with pytest.raises(DuplicateIDError):
        validate_unique_assembly_ids(assemblies)



from collections.abc import Mapping
from typing import Any, Type


def construct_repr(class_obj: Type, fields: Mapping[str, Any]) -> str:
    """Construct a repr string for a class with given fields."""
    field_strs = [f'{key}={value!r}' for key, value in fields.items()]
    field_str = ', '.join(field_strs)
    return f'<{class_obj.__name__} {field_str}>'



import pandas as pd
import pytest

from nasap_net.io import save_reactions
from nasap_net.models import Assembly, BindingSite, Bond, Component, Reaction


@pytest.fixture
def MX2_and_free_L():
    M = Component(kind='M', sites=[0, 1])
    L = Component(kind='L', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MX2 = Assembly(
        id_='MX2',
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]
    )
    free_L = Assembly(id_='free_L', components={'L0': L}, bonds=[])
    MLX = Assembly(
        id_='MLX',
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[],
    )
    free_X = Assembly(id_='free_X', components={'X0': X}, bonds=[])

    return Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 0),
        duplicate_count=4,
        id_='R1',
    )


def test_basic(MX2_and_free_L, tmp_path):
    reactions = [MX2_and_free_L]

    output_file = tmp_path / 'reactions.csv'
    save_reactions(reactions, output_file)

    assert output_file.exists()
    df = pd.read_csv(output_file)
    assert len(df) == 1
    row = df.iloc[0]
    assert row['init_assem_id'] == 'MX2'
    assert row['entering_assem_id'] == 'free_L'
    assert row['product_assem_id'] == 'MLX'
    assert row['leaving_assem_id'] == 'free_X'
    assert row['metal_bs_component'] == 'M0'
    assert row['metal_bs_site'] == 0
    assert row['leaving_bs_component'] == 'X0'
    assert row['leaving_bs_site'] == 0
    assert row['entering_bs_component'] == 'L0'
    assert row['entering_bs_site'] == 0
    assert row['duplicate_count'] == 4
    assert row['id'] == 'R1'


def test_no_id(MX2_and_free_L, tmp_path):
    reaction_no_id = MX2_and_free_L.copy_with(id_=None)
    reactions = [reaction_no_id]

    output_file = tmp_path / 'reactions.csv'
    save_reactions(reactions, output_file)

    assert output_file.exists()
    df = pd.read_csv(output_file)
    assert len(df) == 1
    row = df.iloc[0]
    assert pd.isna(row['id'])



from copy import deepcopy

import pytest

from nasap_net.exceptions import IDNotSetError
from nasap_net.models import Assembly, BindingSite, Component, Reaction


@pytest.fixture
def MX2_plus_free_L():
    M = Component(kind='M', sites=[0, 1])
    L = Component(kind='L', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MX2 = Assembly(
        id_='MX2',
        components={'X0': X, 'M0': M, 'X1': X},
        bonds=[],
    )
    free_L = Assembly(id_='free_L', components={'L0': L}, bonds=[])
    MLX = Assembly(
        id_='MLX',
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[],
    )
    free_X = Assembly(id_='free_X', components={'X0': X}, bonds=[])

    return Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 0),
        duplicate_count=4,
    )


def test___repr__(MX2_plus_free_L):
    repr_str = repr(MX2_plus_free_L)
    assert repr_str == '<Reaction MX2 + free_L -> MLX + free_X>'


def test___repr___with_id(MX2_plus_free_L):
    reaction = MX2_plus_free_L.copy_with(id_='R1')
    repr_str = repr(reaction)
    assert repr_str == '<Reaction ID=R1 MX2 + free_L -> MLX + free_X>'


def test_equation_str(MX2_plus_free_L):
    equation = MX2_plus_free_L.equation_str
    assert equation == 'MX2 + free_L -> MLX + free_X'


def test___str__(MX2_plus_free_L):
    str_repr = str(MX2_plus_free_L)
    assert str_repr == 'MX2 + free_L -> MLX + free_X (x4)'


def test_id_property(MX2_plus_free_L):
    # Error when ID is not set
    with pytest.raises(IDNotSetError):
        _ = MX2_plus_free_L.id_

    # No error when ID is set
    reaction_with_id = MX2_plus_free_L.copy_with(id_='R1')
    assert reaction_with_id.id_ == 'R1'


def test_id_or_none_property(MX2_plus_free_L):
    assert MX2_plus_free_L.id_or_none is None

    reaction_with_id = MX2_plus_free_L.copy_with(id_='R1')
    assert reaction_with_id.id_or_none == 'R1'


def test_copy_with():
    original = Reaction(
        init_assem=Assembly(id_='init_assem_1', components={}, bonds=[]),
        entering_assem=Assembly(id_='entering_assem_1', components={}, bonds=[]),
        product_assem=Assembly(id_='product_assem_1', components={}, bonds=[]),
        leaving_assem=Assembly(id_='leaving_assem_1', components={}, bonds=[]),
        metal_bs=BindingSite('metal_1', 0),
        leaving_bs=BindingSite('leaving_1', 0),
        entering_bs=BindingSite('entering_1', 0),
        duplicate_count=1,
        id_='R_original',
    )

    original_copy = deepcopy(original)

    modified = original.copy_with(
        init_assem=Assembly(id_='init_assem_2', components={}, bonds=[]),
        entering_assem=Assembly(id_='entering_assem_2', components={},
                                bonds=[]),
        product_assem=Assembly(id_='product_assem_2', components={}, bonds=[]),
        leaving_assem=Assembly(id_='leaving_assem_2', components={}, bonds=[]),
        metal_bs=BindingSite('metal_2', 0),
        leaving_bs=BindingSite('leaving_2', 0),
        entering_bs=BindingSite('entering_2', 0),
        duplicate_count=2,
        id_='R_modified',
    )

    assert modified.init_assem == Assembly(id_='init_assem_2', components={}, bonds=[])
    assert modified.entering_assem == Assembly(id_='entering_assem_2', components={}, bonds=[])
    assert modified.product_assem == Assembly(id_='product_assem_2', components={}, bonds=[])
    assert modified.leaving_assem == Assembly(id_='leaving_assem_2', components={}, bonds=[])
    assert modified.metal_bs == BindingSite('metal_2', 0)
    assert modified.leaving_bs == BindingSite('leaving_2', 0)
    assert modified.entering_bs == BindingSite('entering_2', 0)
    assert modified.duplicate_count == 2
    assert modified.id_ == 'R_modified'

    # Ensure original is unchanged
    assert original.init_assem == original_copy.init_assem
    assert original.entering_assem == original_copy.entering_assem
    assert original.product_assem == original_copy.product_assem
    assert original.leaving_assem == original_copy.leaving_assem
    assert original.metal_bs == original_copy.metal_bs
    assert original.leaving_bs == original_copy.leaving_bs
    assert original.entering_bs == original_copy.entering_bs
    assert original.duplicate_count == original_copy.duplicate_count
    assert original.id_ == 'R_original'


def test_copy_with_no_changes(MX2_plus_free_L):
    copied = MX2_plus_free_L.copy_with()

    assert copied.init_assem == MX2_plus_free_L.init_assem
    assert copied.entering_assem == MX2_plus_free_L.entering_assem
    assert copied.product_assem == MX2_plus_free_L.product_assem
    assert copied.leaving_assem == MX2_plus_free_L.leaving_assem
    assert copied.metal_bs == MX2_plus_free_L.metal_bs
    assert copied.leaving_bs == MX2_plus_free_L.leaving_bs
    assert copied.entering_bs == MX2_plus_free_L.entering_bs
    assert copied.duplicate_count == MX2_plus_free_L.duplicate_count
    assert copied.id_or_none is None  # ID should not be copied by default


def test_override_with_none():
    reaction = Reaction(
        init_assem=Assembly(id_='init_assem', components={}, bonds=[]),
        entering_assem=Assembly(id_='entering_assem', components={}, bonds=[]),
        product_assem=Assembly(id_='product_assem', components={}, bonds=[]),
        leaving_assem=Assembly(id_='leaving_assem', components={}, bonds=[]),
        metal_bs=BindingSite('metal', 0),
        leaving_bs=BindingSite('leaving', 0),
        entering_bs=BindingSite('entering', 0),
        duplicate_count=1,
    )
    assert reaction.entering_assem is not None
    assert reaction.leaving_assem is not None

    # Overridden to None when None is explicitly provided
    modified = reaction.copy_with(
        entering_assem=None,
        leaving_assem=None,
    )
    assert modified.entering_assem is None
    assert modified.leaving_assem is None

    # Original value is used when None is not explicitly provided
    unmodified = reaction.copy_with()
    assert unmodified.entering_assem == reaction.entering_assem
    assert unmodified.leaving_assem == reaction.leaving_assem

    # Other fields cannot be set to None even when None is provided
    with pytest.raises(TypeError):
        reaction.copy_with(init_assem=None)  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        reaction.copy_with(product_assem=None)  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        reaction.copy_with(metal_bs=None)  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        reaction.copy_with(leaving_bs=None)  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        reaction.copy_with(entering_bs=None)  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        reaction.copy_with(duplicate_count=None)  # type: ignore[arg-type]
    with pytest.raises(ValueError):
        reaction.copy_with(duplicate_count=0)  # Should be positive integer



from .component_dump import dump_components
from .component_load import load_components
from .semi_light_assembly_dump import dump_semi_light_assemblies
from .semi_light_assembly_load import load_semi_light_assemblies



from collections import Counter
from collections.abc import Sequence

from nasap_net.models import Assembly


def generate_composition_formula(
        assembly: Assembly,
        *,
        order: Sequence[str] | None = None,
) -> str:
    """Generate a composition formula string for the given assembly.

    You can specify the order of component kinds in the formula.
    Components not specified in `comp_kind_order` will appear at the end of
    the formula sorted alphabetically.

    Parameters
    ----------
    assembly : Assembly
        The assembly for which to generate the composition formula.
    order : Sequence[str]
        The order of component kinds to use in the formula.

    Returns
    -------
    str
        The composition formula string.
    """
    kind_counter: Counter[str] = Counter(
        comp.kind for comp in assembly.components.values()
    )
    if order is None:
        full_kind_order = sorted(kind_counter.keys())
    else:
        unspecified_kinds = kind_counter.keys() - set(order)
        sorted_unspecified_kinds = sorted(unspecified_kinds)
        full_kind_order = list(order) + sorted_unspecified_kinds

    formula_parts = []
    for kind in full_kind_order:
        if kind not in kind_counter:
            continue
        count = kind_counter[kind]
        formula_parts.append(_composition_to_str(kind, count))

    return ''.join(formula_parts)


def _composition_to_str(
        kind: str,
        count: int,
) -> str:
    """Convert a component kind and its count to a string representation."""
    if count == 1:
        return kind
    return f'{kind}{count}'



from enum import Enum


class Column(Enum):
    """Column names for reaction data."""
    INIT_ASSEM_ID = 'init_assem_id'
    ENTERING_ASSEM_ID = 'entering_assem_id'
    PRODUCT_ASSEM_ID = 'product_assem_id'
    LEAVING_ASSEM_ID = 'leaving_assem_id'
    METAL_BS_COMPONENT = 'metal_bs_component'
    METAL_BS_SITE = 'metal_bs_site'
    LEAVING_BS_COMPONENT = 'leaving_bs_component'
    LEAVING_BS_SITE = 'leaving_bs_site'
    ENTERING_BS_COMPONENT = 'entering_bs_component'
    ENTERING_BS_SITE = 'entering_bs_site'
    DUPLICATE_COUNT = 'duplicate_count'
    ID_ = 'id'



from nasap_net.models import Assembly, Component
from nasap_net.utils import deduplicate_ids


def test():
    X = Component(kind='X', sites=[0])

    assemblies = [
        Assembly(id_='freeX', components={'X0': X}, bonds=[]),
        Assembly(id_='freeX', components={'X1': X}, bonds=[]),
    ]

    deduplicated = deduplicate_ids(assemblies)

    assert deduplicated == [
        Assembly(id_='freeX', components={'X0': X}, bonds=[]),
        Assembly(id_='freeX_2', components={'X1': X}, bonds=[]),
    ]


def test_keep_order():
    X = Component(kind='X', sites=[0])

    assemblies = [
        Assembly(id_='freeX', components={'X0': X}, bonds=[]),
        Assembly(id_='freeX', components={'X2': X}, bonds=[]),
        Assembly(id_='freeX', components={'X1': X}, bonds=[]),
    ]

    deduplicated = deduplicate_ids(assemblies)

    assert deduplicated == [
        Assembly(id_='freeX', components={'X0': X}, bonds=[]),
        Assembly(id_='freeX_2', components={'X2': X}, bonds=[]),
        Assembly(id_='freeX_3', components={'X1': X}, bonds=[]),
    ]
